{% extends "layout.html" %}
{% set title = 'Ruffus' %}


{% block extrahead %}
<meta content='Ruffus is a Computation Pipeline library for python. It is open-sourced, powerful and user-friendly, and widely used in science and bioinformatics' name='description'/>
<meta content='python, pipeline, computation, cluster, Ruffus, bioinformatics, parallel' name='keywords'/>
<meta content='Leo Goodstadt' name='Author'/>
<meta content='Leo Goodstadt' name='Publisher'/>
<meta content='ruffus@llew.org.uk' name='Email'/>
<meta content='index, follow' name='robots'/>
<meta content='en-uk' name='language'/>
<meta content='United Kingdom' name='country'/>
<meta name='copyright' content='Leo Goodstadt'/>
{% endblock %}


{% block body %}

<p>
<a href="why_ruffus.html#why-ruffus"> <img src="_images/logo.jpg" alt="logo"></a>
</p>


Ruffus is a Computation Pipeline library for python. It is open-sourced, 
powerful and user-friendly, and widely used in science and bioinformatics.<br/><br/>

<h1>Citation:</h1>
<blockquote>
Please cite <I>Ruffus</I> as:<br/>
<div><a class="reference external" href="http://bioinformatics.oxfordjournals.org/content/early/2010/09/16/bioinformatics.btq524">Leo Goodstadt (2010)</a> : 
<strong>Ruffus: a lightweight Python library for computational pipelines.</strong> <em>Bioinformatics</em> 26(21): 2778-2779</div></blockquote><br/>


<h1>Welcome</h1>

<table class="Introduction" align="center" style="margin-left: 30px">
<tr>
  <td width="50%">
    <p><I>Ruffus</I> is designed to allow scientific and other analyses to 
    be automated with the minimum of fuss and the least effort.</p>
    
    These are <I>Ruffus</I>'s strengths:
    <ul>
      <li><b>Lightweight</b>: Suitable for the simplest of tasks</li>

      <li><b>Scalable</b>: Handles even fiendishly complicated pipelines</a> 
    which would cause <i>make</i> or <i>scons</i> to go cross-eyed and recursive.</li>

    <li><b>Standard python</b>: No "clever magic", no pre-processing.  </li>
    <li><b>Unintrusive</b>: Unambitious, lightweight syntax which tries to do this 
        one small thing well.  </li>
  </ul>
    
    <p>Please join me (email: ruffus_lib at llew.org.uk) in setting the direction of 
    this project if you are interested.
    </p>
    
  </td>
  
  <td width="50%">
        
    <img src="_images/front_page_flowchart.png" alt="flowchart">
    
  </td>
                                                                          
</tr>
</table>



                                                            


<h1> Documentation</h1>

<table class="contentstable" align="center" style="margin-left: 30px">
<tr>
  <td width="50%">
    
    <p class="biglink"><a class="biglink" 
        href="installation.html">Download</a><br/>
    <span class="linkdescr">to install <I>Ruffus</I></span></p>

    <p class="biglink"><a class="biglink" 
        href="tutorials/new_tutorial/introduction.html">Simple Tutorial</a><br/>
    <span class="linkdescr">Start here for a quick introduction to <i>Ruffus</i></span></p>

    <p class="biglink"><a class="biglink" 
        href="tutorials/new_tutorial/manual_contents.html">Manual</a>
        <a class="biglink" href="_downloads/ruffus.pdf">(pdf)</a><br/>
        <span class="linkdescr">for an-depth demonstration of all 
        <I>Ruffus</I> features</span></p>

  </td>
  
  <td width="50%">
        

    <p class="biglink"><a class="biglink" 
        href="contents.html">Table of contents</a><br/>
    <span class="linkdescr">for an complete listing of all the documentation</span></p>
    
    <p class="biglink"><a class="biglink" 
        href="faq.html">Frequently Answered Questions</a><br/>
    <span class="linkdescr">for any common problems, clever solutions from the community</span></p>

    <p class="biglink"><a class="biglink" 
        href="design.html">Design</a><br/>
    <span class="linkdescr">to understand the design of <I>Ruffus</I></span></p>

    <p class="biglink"><a class="biglink" 
        href="cheatsheet.html">Cheat Sheet</a><br/>
    <span class="linkdescr">for Ruffus syntax</span></p>
    </td>
                                                                          
</tr>
</table>




<h2>Get <I>Ruffus</I></h2>

<p>
    <i>Ruffus</i> is available as an <a 
    href="http://peak.telecommunity.com/DevCenter/EasyInstall">easy-install</a>able package on the <a href="http://pypi.python.org/pypi/ruffus">Python Package
    Index</a>.
  </p>
Just run:

<div style="margin-left: 30px" class="highlight-python"><pre>sudo pip install ruffus --upgrade</pre> or
</div>


<div style="margin-left: 30px" class="highlight-python"><pre>easy_install -U ruffus</pre>
</div>
  
<br>  
<p>
The very latest (in development) code can be obtained via <a href="https://code.google.com/p/ruffus/source/checkout"> git </a>:
  <div style="color: #ff0000" >
   <pre>git clone https://bunbun68@code.google.com/p/ruffus/</pre>
</div>
</p>

<h1>Feedback and Getting Involved:</h1>
<ul>
  <li>
    The <I>Ruffus</I> project is hosted with <b><a 
        href="http://code.google.com/p/ruffus">Google 
      Code here.</a></b><br><br>
  </li>
  
  <li>
    <img src="https://groups.google.com/forum/my-groups-color.png"
             height=30 alt="Google Groups"> <b>Subscribe to the <a href="https://groups.google.com/forum/#!forum/ruffus_discuss"><i>ruffus_discuss</i> mailing list </a></b>
    <br>    
    <form action="http://groups.google.com/group/ruffus_discuss/boxsubscribe">
    Email: <input type=text name=email>
    <input type=submit name="sub" value="Subscribe">
    </form><br>
  </li>


  <li>
    <a href="http://groups.google.com/group/ruffus_discuss">Check out the mailing list without subscribing here.</a>
    <br>
  </li>
  
  <li>
    <a href="http://code.google.com/p/ruffus/issues/list">Bugs or feature requests can be posted here.</a>
    <br>
  </li>
  
  
</ul>


{% endblock %}

