{% extends "!layout.html" %}
{% block sidebarsearch %}
    {{ super() }}
    <h3> Quick Reference:</h3>
    <ul>
        <h4><a href="{{ pathto('decorators/decorators') }}#core">Core:</a></h4>   
        <ul>
            <li><a href="{{ pathto('decorators/originate') }}">@originate</a>                           </li>
            <li><a href="{{ pathto('decorators/split') }}">@split</a>                                   </li>
            <li><a href="{{ pathto('decorators/transform') }}">@transform</a>                           </li>
            <li><a href="{{ pathto('decorators/merge') }}">@merge</a>                                   </li>
        </ul>

        <h4><a href="{{ pathto('decorators/decorators') }}#advanced">Advanced</a></h4>   
        <ul>
            <li><a href="{{ pathto('decorators/subdivide') }}">@subdivide</a>                           </li>
            <li><a href="{{ pathto('decorators/transform_ex') }}">@transform (add_inputs) </a>          </li>
            <li><a href="{{ pathto('decorators/collate') }}">@collate</a>                               </li>
            <li><a href="{{ pathto('decorators/collate_ex') }}">@collate (add_inputs)</a>               </li>
            <li><a href="{{ pathto('decorators/graphviz') }}">@graphviz</a>                             </li>
            <li><a href="{{ pathto('decorators/mkdir') }}">@mkdir</a>                                   </li>
            <li><a href="{{ pathto('decorators/follows') }}">@follows / mkdir</a>                       </li>
            <li><a href="{{ pathto('decorators/posttask') }}">@posttask touch_file</a>                  </li>
            <li><a href="{{ pathto('decorators/active_if') }}">@active_if</a>                           </li>
            <li><a href="{{ pathto('decorators/jobs_limit') }}">@jobs_limit</a>                         </li>
        </ul>
    
        <h4><a href="{{ pathto('decorators/decorators') }}#combinatorics">Combinatorial:</a></h4>   
        <ul>
            <li><a href="{{ pathto('decorators/product') }}">@product </a>                              </li>
            <li><a href="{{ pathto('decorators/permutations') }}">@permutations </a>                    </li>
            <li><a href="{{ pathto('decorators/combinations') }}">@combinations </a>                    </li>
            <li><a href="{{ pathto('decorators/combinations_with_replacement') }}">@combinations_ _with_replacement </a>                    </li>
        </ul>

                        
        <h4><a href="{{ pathto('decorators/decorators') }}#esoteric">Esoteric</a></h4>   
        <ul>
            <li><a href="{{ pathto('decorators/files_ex') }}">@files (on the fly)</a>                   </li>
            <li><a href="{{ pathto('decorators/parallel') }}">@parallel</a>                             </li>
            <li><a href="{{ pathto('decorators/check_if_uptodate') }}">@check_if_uptodate</a>           </li>
        </ul>


        <h4 ><a href="{{ pathto('decorators/indicator_objects') }}">Indicator objects</a></h4>
        <h4 >Pipeline functions</h4>
        <ul>
            <li><a href="{{ pathto('pipeline_functions') }}#pipeline-functions-pipeline-run">pipeline_run</a>                       </li>
            <li><a href="{{ pathto('pipeline_functions') }}#pipeline-functions-pipeline-printout">pipeline_printout</a>                      </li>
            <li><a href="{{ pathto('pipeline_functions') }}#pipeline-functions-pipeline-printout-graph">pipeline_printout_graph</a>          </li>
            <li><a href="{{ pathto('pipeline_functions') }}#pipeline-functions-pipeline-get-task-names">pipeline_get_task_names</a>          </li>
            <li><a href="{{ pathto('drmaa_wrapper_functions') }}#drmaa-wrapper-functions-drmaa_wrapper-run_job">drmaa_wrapper.run_job</a>          </li>
        </ul>

         <a href="{{ pathto('todo') }}">Future plans</a>
    </ul>
{% endblock %}

{% block rootrellink %}
        Ruffus v. {{ release }}
            <li><a href="{{ pathto('index') }}">Home</a> | </li>
            <li><a href="{{ pathto('contents') }}">Contents</a> | </li>
            <li><a href="{{ pathto('installation') }}">Install</a>  | </li>
            <li><a href="{{ pathto('tutorials/new_tutorial/introduction') }}">Manual</a>  / </li>
            <li><a href="{{ pathto('tutorials/new_tutorial/manual_contents') }}">(TOC)</a>  | </li>
            <li><a href="{{ pathto('faq') }}">FAQ</a>  | </li>
            <li><a href="{{ pathto('cheatsheet') }}">Cheat sheet</a> | </li>
            <li><a href="{{ pathto('tutorials/new_tutorial/command_line') }}">Command Line</a> | </li>
            <li><a href="{{ pathto('gallery') }}">Gallery</a> |  </li>
            <li><a href="{{ pathto('tutorials/new_syntax') }}">New syntax in v 2.6</a> | </li>
            <li><a href="{{ pathto('history') }}">Latest Changes</a> &raquo; </li>
{% endblock %}




