.. |new_manual.introduction.chapter_num|              replace:: **Chapter 1**
.. |new_manual.transform.chapter_num|                 replace:: **Chapter 2**
.. |new_manual.transform_in_parallel.chapter_num|     replace:: **Chapter 3**
.. |new_manual.originate.chapter_num|                 replace:: **Chapter 4**
.. |new_manual.pipeline_printout.chapter_num|         replace:: **Chapter 5**
.. |new_manual.cmdline.chapter_num|                   replace:: **Chapter 6**
.. |new_manual.pipeline_printout_graph.chapter_num|   replace:: **Chapter 7**
.. |new_manual.output_file_names.chapter_num|         replace:: **Chapter 8**
.. |new_manual.mkdir.chapter_num|                     replace:: **Chapter 9**  
.. |new_manual.checkpointing.chapter_num|             replace:: **Chapter 10** 
.. |new_manual.decorators_compendium.chapter_num|     replace:: **Chapter 11**  
.. |new_manual.split.chapter_num|                     replace:: **Chapter 12**  
.. |new_manual.merge.chapter_num|                     replace:: **Chapter 13** 
.. |new_manual.multiprocessing.chapter_num|           replace:: **Chapter 14**  
.. |new_manual.logging.chapter_num|                   replace:: **Chapter 15**  
.. |new_manual.subdivide_collate.chapter_num|         replace:: **Chapter 16**
.. |new_manual.combinatorics.chapter_num|             replace:: **Chapter 17**
.. |new_manual.active_if.chapter_num|                 replace:: **Chapter 18**
.. |new_manual.posttask.chapter_num|                  replace:: **Chapter 19**
.. |new_manual.inputs.chapter_num|                    replace:: **Chapter 20**  
.. |new_manual.on_the_fly.chapter_num|                replace:: **Chapter 21**  
.. |new_manual.parallel.chapter_num|                  replace:: **Chapter 22**  
.. |new_manual.check_if_uptodate.chapter_num|         replace:: **Chapter 23**  
.. |new_manual.flowchart_colours.chapter_num|         replace:: **Appendix 1**  
.. |new_manual.dependencies.chapter_num|              replace:: **Appendix 2**  
.. |new_manual.exceptions.chapter_num|                replace:: **Appendix 3**  
.. |new_manual.ruffus_names.chapter_num|              replace:: **Appendix 4**  
.. |new_manual.deprecated_files.chapter_num|          replace:: **Appendix 5**  
.. |new_manual.deprecated_files_re.chapter_num|       replace:: **Appendix 6**  

