.. |glob|     replace:: `glob`
.. _glob: http://docs.python.org/library/glob.html

.. meta::
   :description: Ruffus is a Computation Pipeline library for python. It is open-sourced, powerful and user-friendly, and widely used in science and bioinformatics
   :keywords: python, pipeline, computation, cluster, Ruffus, bioinformatics, parallel' name='keywords
   :Author: Leo Goodstadt
   :Publisher: Leo Goodstadt
   :Email: ruffus@llew.org.uk
   :robots: index, follow
   :language: en-uk
   :country: United Kingdom
   :copyright: Leo Goodstadt 2009-2014
