/*
 * Sphinx stylesheet -- default theme
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

@import url("css/theme.css");

/*

body {
    font-family: sans-serif;
    font-size: 100%;
    background-color: #11303d;
    color: #000;
    margin: 0;
    padding: 0;
}

div.document {
    background-color: #1c4e63;
}

div.documentwrapper {
    float: left;
    width: 100%;
}

div.bodywrapper {
    margin: 0 0 0 230px;
}

div.body {
    background-color: #ffffff;
    color: #000000;
    padding: 0 20px 30px 20px;
}

div.footer {
    color: #ffffff;
    width: 100%;
    padding: 9px 0 9px 0;
    text-align: center;
    font-size: 75%;
}

div.footer a {
    color: #ffffff;
    text-decoration: underline;
}

div.related {
    background-color: #133f52;
    line-height: 30px;
    color: #ffffff;
}


div.sphinxsidebar {
}

div.sphinxsidebar h3 {
    font-family: 'Trebuchet MS', sans-serif;
    color: #ffffff;
    font-size: 1.4em;
    font-weight: normal;
    margin: 0;
    padding: 0;
}

div.sphinxsidebar h3 a {
    color: #98dbcc;
    margin: 0;
    padding: 0;
}



div.sphinxsidebar h4 {
    font-family: 'Trebuchet MS', sans-serif;
    color: #ffffff;
    font-size: 1.3em;
    font-weight: normal;
    margin: 5px 0 0 0;
    padding: 0;
}

div.sphinxsidebar p {
    color: #ffffff;
}

div.sphinxsidebar p.topless {
    margin: 5px 10px 10px 10px;
}

div.sphinxsidebar ul {
    font-size: 100%;
    margin: 10px;
    padding: 0;
    color: #ffffff;
}

div.sphinxsidebar a {
    color: #98dbcc;
}

div.sphinxsidebar a em{
    color: #98dbcc;
}

div.sphinxsidebar input {
    border: 1px solid #98dbcc;
    font-family: sans-serif;
    font-size: 1em;
}

/* -- body styles ----------------------------------------------------------- */
/*

a {
    color: #355f7c;
    text-decoration: none;
}

a:hover {
    text-decoration: underline;
}

div.body p, div.body dd, div.body li {
    text-align: justify;
    line-height: 130%;
}

div.body h1,
div.body h2,
div.body h3,
div.body h4,
div.body h5,
div.body h6 {
    font-family: 'Trebuchet MS', sans-serif;
    background-color: #f2f2f2;
    font-weight: normal;
    color: #20435c;
    border-bottom: 1px solid #ccc;
    margin: 20px -20px 10px -20px;
    padding: 3px 0 3px 10px;
}

div.body h1 { margin-top: 0; font-size: 200%; color: #0088FF }
div.body h2 { font-size: 200%; color: #000000 }
div.body h3 { font-size: 140%; color: #800080; margin-left: 0px; margin-top: 40px ;font-style:italic;}        
div.body h4 { font-size: 110%; color: #008000; margin-left: 10px; margin-top: 40px;background-color: #F0F0F0;} 
div.body h5 { font-size: 90%; }
div.body h6 { font-size: 80%; }

div.body h3 cite
{
    font-family: monospace;
    font-weight: normal;
    color: #0000ff;
    font-style: normal;
    font-size: 90%; 
}

div.body h3 em
{
    font-weight: normal;
    color: #000000;
    font-style: normal;
    font-size: 80%; 
}

div.body h2 em
{
    font-family: monospace;
    font-weight: normal;
    color: #800080;
    font-style: normal;
    font-size: 100%; 
}

div.body h2 cite
{
    font-family: monospace;
    font-weight: normal;
    color: #0000ff;
    font-style: italic;
    font-size: 80%; 
}

div.body h2 a
{
    font-family: monospace;
    font-weight: normal;
    color: #0000ff;
    font-style: italic;
    font-size: 80%; 
}

div.body h1 em
{
    font-family: monospace;
    font-weight: normal;
    color: #0088FF;
    font-size: 100%; 
}

dt em
{
    font-family: monospace;
    color: #0000ff;
    font-style: italic;
    font-size: 120%; 
}


a.headerlink {
    color: #c60f0f;
    font-size: 0.8em;
    padding: 0 4px 0 4px;
    text-decoration: none;
}

a.headerlink:hover {
    background-color: #c60f0f;
    color: white;
}

div.body p, div.body dd, div.body li {
    text-align: justify;
    line-height: 130%;
}

div.admonition p.admonition-title + p {
    display: inline;
}

div.note {
    background-color: #eee;
    border: 1px solid #ccc;
}

div.seealso {
    background-color: #ffc;
    border: 1px solid #ff6;
}

div.topic {
    background-color: #eee;
}

div.warning {
    background-color: #ffe4e4;
    border: 1px solid #f66;
}

p.admonition-title {
    display: inline;
}

p.admonition-title:after {
    content: ":";
}

pre {
    padding: 5px;
    background-color: #eeffcc;
    color: #333333;
    line-height: 120%;
    border: 1px solid #ac9;
    border-left: none;
    border-right: none;
}

tt {
    background-color: #ecf0f3;
    padding: 0 1px 0 1px;
    font-size: 0.95em;
}


/*
div.sphinxsidebar h3 {
    font-family: 'Trebuchet MS', sans-serif;
    color: #ffffff;
    font-size: 1.4em;
    font-weight: normal;
    margin: 0 0 0 0;
    padding: 0 0 0 0;
}

div.sphinxsidebar h3 a{
    color: #98dbcc;
    margin: 0px 0px 0px 0px;
    padding: 0px 0px 0px 0px;
}

*/



/*

    Override Pygment style so that I can annotate the code

    The key style are comments within highlighted sections. I have set these to have a
    white background with big bold red text, so that it stands out from and does not appear
    to be part of the surrounding code. This way I don't have to screen shot and highlight
    manually the code in photoshop, and the embedded bitmapped code doesn't get out of date,...
    and the code is selectable, and it fits into a sane workflow etc. etc.

    I use :nth-of-type(n) pseudo class (but one which matches everything)
        so that this is a specialisation of the css .hll class style in pygments.css

    Nasty hack but this way I don't have to modify pygments.css by hand each time, and I don't have
    to create my own pygments style.

*/

.highlight .nd:nth-of-type(n)   { background-color: #ffff88; color: red; font-weight: bold; outline:red dotted thin}/**/
.highlight .hll:nth-of-type(n)   { font-weight: bold}
.highlight .hll .c:nth-of-type(n)    { color: red; font-weight: bold; font-style: normal;font-size:120%} /* Comment */


div.related a {
    font-size: 120%;    /* new */
    color: #c0c0FF;     /* new */
    /*color: #ffffff;     /* orig */
}

/* override table width restrictions */
.wy-table-responsive table td, .wy-table-responsive table th {
    white-space: normal;
}

.wy-table-responsive {
    margin-bottom: 24px;
    max-width: 100%;
    overflow: visible;
}
